#include "stdafx.h"
#include "TemplateMaker.hpp"
#include <iostream>
#include <stdexcept>
#include <boost/shared_ptr.hpp>
#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>
#include <harkio2/harkio.hpp>
#include "utils_harkio2.hpp"
#include "OutputType.hpp"
#include "OutputFormat.hpp"
#include "CommaSeparatedOption.hpp"
#include "Incrementer.hpp"
#include "StringVariableImplanter.hpp"
#include "CommonVariables.hpp"

#include "debug.hpp"




namespace po = boost::program_options;

namespace
{

template <typename T>
boost::shared_ptr<T> MakeBoostSharedPtr(T* p)
{
    return boost::shared_ptr<T>(p);
}








/** @brief StringImplanter向けの変数マップを更新する.
 * AZIMUTH, ELEVATION, HEIGHTの3つを更新する。
 */
void UpdateImplantVariables(std::map<std::string, double>& vm, const harktool4::CommonVariables& vars)
{

    vm["AZIMUTH"]   = *vars.azimuth;
    vm["ELEVATION"] = *vars.elevation;
    vm["HEIGHT"]    = int( round(*vars.height * 1000)); // meter -> millimeter
    vm["RADIUS"]    = int( round(vars.radius * 1000)); //RADIUS対応

}

}




namespace harktool4
{

class TemplateMaker::Impl
{
public:
    Impl(const po::variables_map& vm, std::shared_ptr<std::ostream> output_file)
        : vm_(vm), output_file_(output_file)
    {}

    void Run();

private:
    const po::variables_map& vm_;
    std::shared_ptr<std::ostream> output_file_;

    boost::shared_ptr<harkio2::harkconfig::MicArrayLocationFile2_0>     MakeMic();
    boost::shared_ptr<harkio2::harkconfig::NoiseLocationFile2_0>        MakeNoise();
    boost::shared_ptr<harkio2::harkconfig::ImpulseResponseListFile2_0>  MakeIR();
    boost::shared_ptr<harkio2::harkconfig::TSPListFile2_0>              MakeTSP();
    std::vector<unsigned int> GetMicChannels(size_t nummic);
};





void TemplateMaker::Impl::Run()
{

	DEBUG_PRINT("@ run start: %s[%d]\n", __FILE__ ,__LINE__);


    OutputFormat output_format = OutputFormat::XML;
    if (vm_.count("output-format"))
    {
        output_format = vm_["output-format"].as<OutputFormat>();
    }

    std::unique_ptr<harkio2::harkconfig::HarkConfigFileWriter> writer;
    switch (output_format)
    {
    case OutputFormat::XML:
        writer.reset(new harkio2::harkconfig::HarkConfigFileXMLWriter(*output_file_));
        break;
    default:
        std::cerr << "unsupported output format in template mode" << std::endl;
        return;
    }

    OutputType output_type;
    if (vm_.count("output-type"))
    {
        output_type = vm_["output-type"].as<OutputType>();
    }
    else
    {
        std::cerr << "in template mode, output type must be specified" << std::endl;
        return;
    }

    boost::shared_ptr<harkio2::harkconfig::HarkConfigFile> hc_file;

    try
    {
        switch (output_type)
        {
        case OutputType::Mic:
            hc_file = MakeMic();
            break;
        case OutputType::Noise:
            hc_file = MakeNoise();
            break;
        case OutputType::IR:
            hc_file = MakeIR();
            break;
        case OutputType::TSP:
            hc_file = MakeTSP();
            break;
        default:
            throw std::runtime_error("invalid output type for making template");
        }
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << std::endl;
        return;
    }

    writer->Write(hc_file);
}





boost::shared_ptr<harkio2::harkconfig::MicArrayLocationFile2_0>     TemplateMaker::Impl::MakeMic()
{


	DEBUG_PRINT("@ MakeMic start: %s[%d]\n", __FILE__ ,__LINE__);


    CommonVariables vars(vm_);

    /* nummicオプションが明示的に指定されておらず
     * 代わりにnumaziオプションが明示的に指定されているならば、
     * そちらを優先する
     */
    const unsigned int nummic = (vm_["nummic"].defaulted() && !vm_["numazi"].defaulted()) ?
        (vars.numazi * vars.num_vertical) : vars.nummic;

    std::vector<harkio2::harkconfig::Mic1> mics(0);
    std::vector<unsigned int> mic_channels = GetMicChannels(nummic);
	std::vector<harkio2::harkconfig::Mic1> buff(0);
//	std::cout << "mic_channels.size()=" << mic_channels.size() << " nummic=" << nummic  << std::endl; 

	const unsigned int  new_nummic = mic_channels.size();

	//mic番号チェック
    if (mic_channels.size() > nummic)
    {
        throw std::runtime_error("the mic-channels' size > nummic");
		std::cerr << "mic_channels.size()=" << mic_channels.size() << " nummic=" << nummic << std::endl; 
    }
    mics.resize(new_nummic);
    buff.resize(nummic);



    size_t i = 0;
    while (i < nummic)
    {
        vars.height.Reset();
        vars.elevation.Reset();

        for (size_t h = 0; h < vars.num_vertical && i < nummic; ++h, ++i)
        {
            buff[i].SetMicIndex(i);
            buff[i].SetPosition(MakePosition(vars.use_height, *vars.azimuth, *vars.height, *vars.elevation, vars.radius));
			const auto& pos2 = buff[ i].GetPosition().GetPositionVector();	
//			printf("%d buff:  x=%.4f y=%.4f z=%.4f\n",i, pos2.GetX() , pos2.GetY() , pos2.GetZ());

            ++vars.height;
            ++vars.elevation;
        }
        ++vars.azimuth;
    }


	//mic_channelsのbuffのデータをmicsに設定
	i = 0;
    while (i < new_nummic)
    {
		mics[i].SetMicIndex( buff[ mic_channels[i] ].GetMicIndex() );

		const auto& pos = buff[ mic_channels[i] ].GetPosition().GetPositionVector();	
		mics[i].SetPosition(  harkio2::MakePositionFromVector<float>(harkio2::Vector3<float>(pos.GetX() , pos.GetY() , pos.GetZ() )) );
//		printf("%d mic_channel=%d x=%.4f y=%.4f z=%.4f\n",i, mic_channels[i],pos.GetX()  ,pos.GetY() , pos.GetZ());
		i++;
	}



    auto data = MakeBoostSharedPtr(
            new harkio2::harkconfig::MicArrayLocationFile2_0());

    data->SetMics(mics);
    data->SetComment("created by harktool4");

    return data;

}





boost::shared_ptr<harkio2::harkconfig::NoiseLocationFile2_0>        TemplateMaker::Impl::MakeNoise()
{

	DEBUG_PRINT("@ MakeNoise start: %s[%d]\n", __FILE__ ,__LINE__);


    CommonVariables vars(vm_);

    std::vector<harkio2::harkconfig::NoiseSource1> sources(vars.numazi * vars.num_vertical);

    for (size_t i = 0; i < vars.numazi; ++i)
    {
        vars.height.Reset();
        vars.elevation.Reset();

        for (size_t h = 0; h < vars.num_vertical; ++h)
        {
            sources[i * vars.num_vertical + h].SetNoiseIndex(i);
            sources[i * vars.num_vertical + h].SetPosition(
                    MakePosition(vars.use_height, *vars.azimuth, *vars.height, *vars.elevation, vars.radius));

            ++vars.height;
            ++vars.elevation;
        }
        ++vars.azimuth;
    }

    auto data = MakeBoostSharedPtr(
            new harkio2::harkconfig::NoiseLocationFile2_0());
    data->SetNoiseSources(sources);
    data->SetComment("created by harktool4");

    return data;
}




boost::shared_ptr<harkio2::harkconfig::ImpulseResponseListFile2_0>  TemplateMaker::Impl::MakeIR()
{

	DEBUG_PRINT("@ MakeIR start: %s[%d]\n", __FILE__ ,__LINE__);


    CommonVariables vars(vm_);

    std::vector<unsigned int> mic_channels = GetMicChannels(vars.nummic);
    if (mic_channels.size() != vars.nummic)
    {
        throw std::runtime_error("the mic-channels' size != nummic");
    }

    std::string filepath_template = vm_.count("filepath") ? vm_["filepath"].as<std::string>() : "D{AZIMUTH:%03d}_ch{CHANNEL:%02d}.flt";

    std::vector<harkio2::harkconfig::Source2> sources(vars.numazi * vars.num_vertical);

    // for StringVariableImplanter
    std::map<std::string, double> variables_int;

    // 横方向
    for (size_t i = 0; i < vars.numazi; ++i)
    {
        vars.height.Reset();
        vars.elevation.Reset();

        // 縦方向
        for (size_t h = 0; h < vars.num_vertical; ++h)
        {
            sources[i * vars.num_vertical + h].SetPosition(
                    MakePosition(vars.use_height, *vars.azimuth, *vars.height, *vars.elevation, vars.radius));

            UpdateImplantVariables(variables_int, vars); // AZIMUTH, ELEVATION, HEIGHTを更新
            std::vector<harkio2::harkconfig::File1> files(vars.nummic);
            for (size_t j = 0; j < vars.nummic; ++j)
            {
                variables_int["CHANNEL"] = static_cast<int>(mic_channels[j]);
                files[j].SetName((StringVariableImplanter(filepath_template) % variables_int).str());
                files[j].SetMicIndices({ mic_channels[j] });
            }

            sources[i * vars.num_vertical + h].SetFiles(files);
            sources[i * vars.num_vertical + h].SetCutStart(vars.cut_start);
            sources[i * vars.num_vertical + h].SetCutEnd(vars.cut_end);

            ++vars.height;
            ++vars.elevation;
        }
        ++vars.azimuth;
    }

    auto data = MakeBoostSharedPtr(
            new harkio2::harkconfig::ImpulseResponseListFile2_0());
    data->SetNumMic(vars.nummic);
    data->SetSourceFileType(harkio2::FileType::FLOAT);
    data->SetSources(sources);
    data->SetComment("created by harktool4");

    return data;
}




//--------------------------------------------------------------------/
//
//
//
//--------------------------------------------------------------------/
boost::shared_ptr<harkio2::harkconfig::TSPListFile2_0>   TemplateMaker::Impl::MakeTSP()
{


	DEBUG_PRINT("@ MakeTSP start: %s[%d]\n", __FILE__ ,__LINE__);


    CommonVariables vars(vm_);

    const unsigned int synchronous_average = vm_["synchronous-average"].as<unsigned int>();
    const unsigned int tsp_offset = vm_["tsp-offset"].as<unsigned int>();
    const unsigned int tsp_length = vm_["tsp-length"].as<unsigned int>();
    const unsigned int signal_max = vm_["signal-max"].as<unsigned int>();
    const std::string original_impulse_file = vm_["original-impulse-file"].as<std::string>();

	
	//std::cout << "nummic=" << vars.nummic  << std::endl; 

    std::vector<unsigned int> mic_channels = GetMicChannels(vars.nummic);
    if (mic_channels.size() != vars.nummic)
    {
        throw std::runtime_error("the mic-channels' size != nummic");
    }

    std::string filepath_template = vm_.count("filepath") ? vm_["filepath"].as<std::string>() : "D{AZIMUTH:%03d}_00.wav";

    std::vector<harkio2::harkconfig::Source2> sources(vars.numazi * vars.num_vertical);



	int o_file_type = 0;
	if( original_impulse_file.find(".wav") == std::string::npos)
	{
		o_file_type = 0; // not wav
	}
	else
	{
		o_file_type = -1; // wav
	}
	


    // for StringVariableImplanter
    std::map<std::string, double> variables_int;
    for (size_t i = 0; i < vars.numazi; ++i)
    {
        vars.height.Reset();
        vars.elevation.Reset();

        // 縦方向
        for (size_t h = 0; h < vars.num_vertical; ++h)
        {
            sources[i * vars.num_vertical + h].SetPosition(
                    MakePosition(vars.use_height, *vars.azimuth, *vars.height, *vars.elevation, vars.radius));

            UpdateImplantVariables(variables_int, vars); // AZIMUTH, ELEVATION, HEIGHTを更新

            std::vector<harkio2::harkconfig::File1> files(1);

            files[0].SetName((StringVariableImplanter(filepath_template) % variables_int).str());
            files[0].SetMicIndices(mic_channels);
            sources[i * vars.num_vertical + h].SetFiles(files);
            sources[i * vars.num_vertical + h].SetCutStart(vars.cut_start);
            sources[i * vars.num_vertical + h].SetCutEnd(vars.cut_end);

            ++vars.height;
            ++vars.elevation;
        }
        ++vars.azimuth;
    }

    auto data = MakeBoostSharedPtr(
            new harkio2::harkconfig::TSPListFile2_0());
    data->SetNumMic(vars.nummic);
    data->SetSourceFileType(harkio2::FileType::FLOAT);
    data->SetSynchronousAverage(synchronous_average);
    data->SetTSPOffset(tsp_offset);
    data->SetTSPLength(tsp_length);
    data->SetSignalMax(signal_max);

	if(o_file_type == 0 )
	{
	    data->SetOriginalImpulseFile(harkio2::TypedFilename(original_impulse_file, harkio2::FileType::FLOAT));
	}
	else
	{
	    data->SetOriginalImpulseFile(harkio2::TypedFilename(original_impulse_file, harkio2::FileType::WAVE));
	}
	

    data->SetSources(sources);
    data->SetComment("created by harktool4");

    return data;
}





/** @brief マイクチャンネルのリストを取得または生成して返す.
 * mic-channelsコマンドラインオプションが指定されていれば、
 * その引数を解析して返す。
 * 指定されていなければ、{ 0, 1, .. , nummic-1 }という配列を生成して返す。
 * @param nummic mic-channelsが指定されていない場合のチャンネル数
 */
std::vector<unsigned int> TemplateMaker::Impl::GetMicChannels(size_t nummic)
{

	DEBUG_PRINT("@ GetMicChannels start: %s[%d]\n", __FILE__ ,__LINE__);


    if (vm_.count("mic-channels"))
    {
        // コマンドラインでmic-channelsが指定されていれば、カンマで区切って読み込む
        return vm_["mic-channels"].as<CommaSeparatedOption<unsigned int>>().values;
    }
    else
    {
        // mic-channelsが指定されていないなら、
        // { 0, 1, 2, ... , nummic-1 } という配列を作って返す
        std::vector<unsigned int> mic_channels(nummic);
        for (size_t i = 0; i < nummic; ++i) { mic_channels[i] = i; }
        return mic_channels;
    }
}





TemplateMaker::TemplateMaker(const boost::program_options::variables_map& vm, std::shared_ptr<std::ostream> output_file)
    : impl_(new Impl(vm, output_file))
{}

void TemplateMaker::Run()
{
    impl_->Run();
}

}

